/*****************************************************************************
 * atmo.c : "Atmo Light Filter" video filter
 *****************************************************************************
 * Copyright (C) 2000-2006 the VideoLAN team
 * $Id$
 *
 * Authors: Andr Weber (WeberAndre@gmx.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include <stdlib.h>                                      /* malloc(), free() */
#include <string.h>
#include <math.h>                                            /* sin(), cos() */

#include <vlc/vlc.h>
#include <vlc/vout.h>
#include "filter_common.h"
#include "vlc_image.h"


/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static int  CreateFilter    ( vlc_object_t * );
static void DestroyFilter   ( vlc_object_t * );

static int InitFilter( vout_thread_t *p_vout );
static void EndFilter( vout_thread_t *p_vout );

static int  SendEvents   ( vlc_object_t *, char const *, vlc_value_t, vlc_value_t, void * );
static void Render( vout_thread_t *p_vout, picture_t *p_pic );
static int Control( vout_thread_t *p_vout, int i_query, va_list args );


#define lvsGDI	         0
#define lvsExternal      1

#define cemDisabled	     0
#define	cemStaticColor   1
#define	cemLivePicture   2
#define cemColorChange   3
#define	cemLrColorChange 4


#define MODE_TEXT N_("Atmo Mode after Closing VLC")
#define MODE_LONGTEXT N_("Defines the effect after closing VLC")


/*****************************************************************************
 * Module descriptor
 *****************************************************************************/
vlc_module_begin();
    set_description( _("AtmoLight filter") );
    set_shortname( _( "atmo" ));
    set_capability( "video filter", 0 );
    set_category( CAT_VIDEO );
    set_subcategory( SUBCAT_VIDEO_VFILTER );

    add_integer( "atmo-mode", 0, NULL, MODE_TEXT, MODE_LONGTEXT, VLC_FALSE );

    add_shortcut( "atmo" );
    set_callbacks( CreateFilter, DestroyFilter  );
vlc_module_end();


/***
 * Function Prototypes for COM Wrapper DLL because I wasn't able to get the COM Stuffed compiled
 * below cygwin and MingW -- so moved all COM call to another DLL which could be compiled with
 * Visual Studio...
*/
typedef int (TAtmoInitialize(void));
typedef void (TAtmoFinalize(void));
typedef int (TAtmoSwitchEffect(int newMode));
typedef int (TAtmoSetLiveSource(int newSource));
typedef void (TAtmoCreateTransferBuffers(int FourCC,int bytePerPixel,int width,int height));
typedef uint8_t* (TAtmoLockTransferBuffer(void));
typedef void (TAtmoSendPixelData(void));

/*****************************************************************************
 * vout_sys_t: AtmoLight output method descriptor (my global variables for my thread:-))
 *****************************************************************************
 * This structure is part of the video output thread descriptor.
 * It describes the AtmoLight specific properties of an output thread.
 *****************************************************************************/
struct vout_sys_t
{
    vout_thread_t *p_vout;
    int i_atmo_mode;

    int i_enabled;
    int i_AtmoOldEffect;

    int i_atmo_width;
    int i_atmo_height;

    // External Library as wrapper arround my COM Stuff
    HINSTANCE h_AtmoCtrl;
    TAtmoInitialize *AtmoInitialize;
    TAtmoFinalize *AtmoFinalize;
    TAtmoSwitchEffect *AtmoSwitchEffect;
    TAtmoSetLiveSource *AtmoSetLiveSource;
    TAtmoCreateTransferBuffers *AtmoCreateTransferBuffers;
    TAtmoLockTransferBuffer *AtmoLockTransferBuffer;
    TAtmoSendPixelData *AtmoSendPixelData;
};


/*****************************************************************************
 * Create: allocates AtmoLight video thread output method
 *****************************************************************************
 * This function allocates and initializes a AtmoLight vout method.
 *****************************************************************************/
static int CreateFilter( vlc_object_t *p_this )
{
    vout_thread_t *p_vout = (vout_thread_t *)p_this;
    vout_sys_t *p_sys;

    /* Allocate structure */
    p_sys = malloc( sizeof( vout_sys_t ) );;
    p_vout->p_sys = p_sys;
    if( p_vout->p_sys == NULL )
    {
        msg_Err( p_vout, "out of memory for p_sys structure" );
        return VLC_ENOMEM;
    }

    /* default filter disabled until DLL loaded and Init Success!*/
    p_sys->i_enabled     = 0;
    /* setup default mini image size (may be later a user option) */
    p_sys->i_atmo_width  = 64;
    p_sys->i_atmo_height = 48;

    /* Load the Com Wrapper Library (source available) */
    p_sys->h_AtmoCtrl = LoadLibraryA("AtmoCtrlLib.dll");
    if(p_sys->h_AtmoCtrl != NULL) {
       msg_Dbg( p_vout, "LoadLibrary('AtmoCtrlLib.dll'); Success");  

       /* importing all required functions I hope*/
       p_sys->AtmoInitialize            = (TAtmoInitialize *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoInitialize");
       if(!p_sys->AtmoInitialize)
          msg_Err( p_vout, "export AtmoInitialize missing."); 

       p_sys->AtmoFinalize              = (TAtmoFinalize *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoFinalize");
       if(!p_sys->AtmoFinalize)
          msg_Err( p_vout, "export AtmoFinalize missing."); 

       p_sys->AtmoSwitchEffect          = (TAtmoSwitchEffect *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoSwitchEffect");
       if(!p_sys->AtmoSwitchEffect)
          msg_Err( p_vout, "export AtmoSwitchEffect missing."); 

       p_sys->AtmoSetLiveSource         = (TAtmoSetLiveSource *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoSetLiveSource");
       if(!p_sys->AtmoSetLiveSource)
          msg_Err( p_vout, "export AtmoSetLiveSource missing."); 

       p_sys->AtmoCreateTransferBuffers = (TAtmoCreateTransferBuffers *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoCreateTransferBuffers");
       if(!p_sys->AtmoCreateTransferBuffers)
          msg_Err( p_vout, "export AtmoCreateTransferBuffers missing."); 

       p_sys->AtmoLockTransferBuffer    = (TAtmoLockTransferBuffer *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoLockTransferBuffer");
       if(!p_sys->AtmoLockTransferBuffer)
          msg_Err( p_vout, "export AtmoLockTransferBuffer missing."); 

       p_sys->AtmoSendPixelData         = (TAtmoSendPixelData *)GetProcAddress(p_sys->h_AtmoCtrl,"AtmoSendPixelData");
       if(!p_sys->AtmoSendPixelData)
          msg_Err( p_vout, "export AtmoSendPixelData missing."); 
    }

    /* some Test variable..*/
    p_sys->i_atmo_mode = config_GetInt( p_vout, "atmo-mode" );
   
    /* further Setup Function pointers for videolan for calling my filter */
    p_vout->pf_init = InitFilter;
    p_vout->pf_end = EndFilter;
    p_vout->pf_manage = NULL;
    p_vout->pf_render = Render;
    p_vout->pf_display = NULL;
    p_vout->pf_control = Control;
    
    return VLC_SUCCESS;
}


/*****************************************************************************
 * DestroyFilter: destroy AtmoLight video thread output method
 *****************************************************************************
 * Terminate an output method created by CreateFilter
 *****************************************************************************/
static void DestroyFilter( vlc_object_t *p_this )
{
    vout_thread_t *p_vout = (vout_thread_t *)p_this;
    vout_sys_t *p_sys =  p_vout->p_sys;

    if( p_sys->p_vout )
    {
        DEL_CALLBACKS( p_sys->p_vout, SendEvents );
        vlc_object_detach( p_sys->p_vout );
        vout_Destroy( p_sys->p_vout );
    }
    DEL_PARENT_CALLBACKS( SendEventsToChild );

    if(p_sys->h_AtmoCtrl!=NULL) {
       FreeLibrary(p_sys->h_AtmoCtrl);
    }

    free( p_sys );
}


/*****************************************************************************
 * Init: initialize AtmoLight video thread output method
 *****************************************************************************/
static int InitFilter( vout_thread_t *p_vout )
{
    int i_index;
    picture_t *p_pic;
    video_format_t fmt = {0};
    char buffer[200];

    vout_sys_t *p_sys = p_vout->p_sys;

    I_OUTPUTPICTURES = 0;

    /* Initialize the output structure */
    p_vout->output.i_chroma = p_vout->render.i_chroma;
    p_vout->output.i_width  = p_vout->render.i_width;
    p_vout->output.i_height = p_vout->render.i_height;
    p_vout->output.i_aspect = p_vout->render.i_aspect;
    p_vout->fmt_out = p_vout->fmt_in;
    fmt = p_vout->fmt_out;

    /* Try to open the real video output */
    msg_Dbg( p_vout, "InitFilter AtmoLight Filter Thread" );

    p_sys->p_vout = vout_Create( p_vout, &fmt );
    /* Everything failed */
    if( p_sys->p_vout == NULL ) {
        msg_Err( p_vout, "cannot open vout for AtmoLight, aborting" );
        return VLC_EGENERIC;
    }

    ALLOCATE_DIRECTBUFFERS( VOUT_MAX_PICTURES );
    ADD_CALLBACKS( p_vout->p_sys->p_vout, SendEvents );
    ADD_PARENT_CALLBACKS( SendEventsToChild );

    /*
       if my DLL was loaded successfully Initialize it now - in this thread (because COM doesn't like it
       very much if objects are used in different threads ;-) sucksz
    */
    if(p_sys->h_AtmoCtrl != NULL) {
       if(p_sys->AtmoInitialize!=NULL) {
          msg_Dbg( p_vout, "call AtmoInitialize()");    
          if(p_sys->AtmoInitialize() == 1) {
              msg_Dbg( p_vout, "AtmoInitialize Ok!");
              /* Setup Transferbuffers for 64 x 48 , RGB with 32bit Per Pixel */
              p_sys->AtmoCreateTransferBuffers(BI_RGB, 4, p_sys->i_atmo_width, p_sys->i_atmo_height);

              /* say the userspace driver that a live mode should be activated 
                 the functions returns the old mode for later restore!
              */
              p_sys->i_AtmoOldEffect = p_sys->AtmoSwitchEffect(cemLivePicture);

              /* live view can have two differnt source the AtmoWinA internal GDI Screencapture
                 and the external one - which we need here...
              */
              p_sys->AtmoSetLiveSource(lvsExternal);

              p_sys->i_enabled = 1; 
           }  
       }
    }
    return VLC_SUCCESS;
}


/*****************************************************************************
 * End: terminate Distort video thread output method
 *****************************************************************************/
static void EndFilter( vout_thread_t *p_vout )
{
    int i_index;
    vout_sys_t *p_sys = p_vout->p_sys;

    /* Free the fake output buffers we allocated */
    for( i_index = I_OUTPUTPICTURES ; i_index ; )
    {
        i_index--;
        free( PP_OUTPUTPICTURE[ i_index ]->p_data_orig );
    }

    if(p_sys->h_AtmoCtrl!=NULL) {
       if(p_sys->i_enabled == 1) {
           msg_Dbg( p_vout, "EndFilter Switchback to AtmoMode %d", p_sys->i_AtmoOldEffect);

           /* restore AtmoWinA Mode .. to old effect before VideoLan! */

          if(p_sys->i_AtmoOldEffect!=cemLivePicture)
             p_sys->AtmoSwitchEffect(p_sys->i_AtmoOldEffect);

          if(p_sys->i_AtmoOldEffect==cemLivePicture)
             p_sys->AtmoSetLiveSource(lvsGDI);

          /* cleanup Library Globals VarArrays, Global COM Object pointers (Release them) */
          p_sys->AtmoFinalize();

          /* disable filter method .. */
          p_sys->i_enabled = 0;
       }
    }

}


/*
 function stolen from some other videolan source filter ;-)
 for the moment RGB is OK... but better would be a direct transformation
 from YUV --> HSV 
*/
static inline void yuv_to_rgb( uint8_t *r, uint8_t *g, uint8_t *b,
                               uint8_t y1, uint8_t u1, uint8_t v1 )
{
    /* macros used for YUV pixel conversions */
#   define SCALEBITS 10
#   define ONE_HALF  (1 << (SCALEBITS - 1))
#   define FIX(x)    ((int) ((x) * (1<<SCALEBITS) + 0.5))
#   define CLAMP( x ) (((x) > 255) ? 255 : ((x) < 0) ? 0 : (x));

    int y, cb, cr, r_add, g_add, b_add;

    cb = u1 - 128;
    cr = v1 - 128;
    r_add = FIX(1.40200*255.0/224.0) * cr + ONE_HALF;
    g_add = - FIX(0.34414*255.0/224.0) * cb
            - FIX(0.71414*255.0/224.0) * cr + ONE_HALF;
    b_add = FIX(1.77200*255.0/224.0) * cb + ONE_HALF;
    y = (y1 - 16) * FIX(255.0/219.0);
    *r = CLAMP((y + r_add) >> SCALEBITS);
    *g = CLAMP((y + g_add) >> SCALEBITS);
    *b = CLAMP((y + b_add) >> SCALEBITS);
}

/*
  CreateMiniImage extracts a 64x48 Pixel Image from the Frame
  (there is a small border arround thats why the loops starts with one)
  without interpolation 

*/
static void CreateMiniImage( vout_thread_t *p_vout, picture_t *p_inpic)
{
    vout_sys_t *p_sys = p_vout->p_sys;
    int i_col;
    int i_row;
    uint8_t *p_src_y;
    uint8_t *p_src_u;
    uint8_t *p_src_v;
    uint8_t *p_transfer;
    uint8_t *p_rgb_dst_line_red;
    uint8_t *p_rgb_dst_line_green;
    uint8_t *p_rgb_dst_line_blue;
    int i_xpos_y;
    int i_xpos_u;
    int i_xpos_v;
    
    /*
      Lock the before created VarArray (AtmoCreateTransferBuffers) inside my Wrapper library and give 
      me a pointer to the buffer!
    */
    p_transfer = (uint8_t *)p_sys->AtmoLockTransferBuffer();
    if(p_transfer == NULL) {
       msg_Err( p_vout, "AtmoLight locking transferbuffer failed. AtmoLight will be disabled!"); 
       p_sys->i_enabled = 0;
       return;
    }

    /* calcute Pointers for Storage of R G B (A) */
    p_rgb_dst_line_red   = p_transfer;
    p_rgb_dst_line_green = p_rgb_dst_line_red;
    p_rgb_dst_line_green++;
    p_rgb_dst_line_blue  = p_rgb_dst_line_green;
    p_rgb_dst_line_blue++;

    int i_row_count = p_sys->i_atmo_height + 1;
    int i_col_count = p_sys->i_atmo_width + 1;

    /*  these two ugly loops extract the small image - goes it faster? how?
        the loops are so designed that there is a small border around the extracted image
        so we wont get column and row - zero from the frame, and not the most right 
        and bottom pixels --- which may be clipped on computers useing TV out - through
        overscan!
    */
    for(i_row = 1; i_row < i_row_count; i_row++) {
        // calcute the current Lines in the source planes for this outputrow
        p_src_y = p_inpic->p[Y_PLANE].p_pixels + p_inpic->p[Y_PLANE].i_pitch * ((i_row * p_inpic->p[Y_PLANE].i_visible_lines)/i_row_count);
        p_src_u = p_inpic->p[U_PLANE].p_pixels + p_inpic->p[U_PLANE].i_pitch * ((i_row * p_inpic->p[U_PLANE].i_visible_lines)/i_row_count);
        p_src_v = p_inpic->p[V_PLANE].p_pixels + p_inpic->p[V_PLANE].i_pitch * ((i_row * p_inpic->p[V_PLANE].i_visible_lines)/i_row_count);

        for(i_col = 1; i_col < i_col_count; i_col++) {
            i_xpos_y = (p_inpic->p[Y_PLANE].i_visible_pitch * i_col) / i_col_count; 
            i_xpos_u = (p_inpic->p[U_PLANE].i_visible_pitch * i_col) / i_col_count; 
            i_xpos_v = (p_inpic->p[V_PLANE].i_visible_pitch * i_col) / i_col_count; 

            yuv_to_rgb(p_rgb_dst_line_red,p_rgb_dst_line_green,p_rgb_dst_line_blue, p_src_y[i_xpos_y], p_src_u[i_xpos_u], p_src_v[i_xpos_v]);
            
            /* +4 because Output Image should be RGB32 with Dword Alignment! */
            p_rgb_dst_line_red+=4;            
            p_rgb_dst_line_green+=4;
            p_rgb_dst_line_blue+=4;
        }
    }
    p_sys->AtmoSendPixelData();

}


/*****************************************************************************
 * SendEventsToChild: forward events to the child/children vout
 *****************************************************************************/
static int SendEventsToChild( vlc_object_t *p_this, char const *psz_var,
                       vlc_value_t oldval, vlc_value_t newval, void *p_data )
{
    vout_thread_t *p_vout = (vout_thread_t *)p_this;
    var_Set( p_vout->p_sys->p_vout, psz_var, newval );
    return VLC_SUCCESS;
}


/*****************************************************************************
 * Render: displays previously rendered output
 *****************************************************************************
 * This function send the currently rendered image to Distort image, waits
 * until it is displayed and switch the two rendering buffers, preparing next
 * frame.
 *****************************************************************************/
static void Render( vout_thread_t *p_vout, picture_t *p_pic )
{
    picture_t *p_outpic;
    vout_sys_t *p_sys = p_vout->p_sys;

    /* This is a new frame. Get a structure from the video_output. */
    while( ( p_outpic = vout_CreatePicture( p_sys->p_vout, 0, 0, 0 ) ) == NULL )
    {
        if( p_vout->b_die || p_vout->b_error )
        {
            return;
        }
        msleep( VOUT_OUTMEM_SLEEP );
    }
    vout_DatePicture( p_sys->p_vout, p_outpic, p_pic->date );
    if(p_sys->i_enabled == 1) {
       CreateMiniImage(p_vout, p_pic);
    }
    vout_CopyPicture( p_vout, p_outpic, p_pic );

    vout_DisplayPicture( p_sys->p_vout, p_outpic );
}

/*****************************************************************************
 * Control: control facility for the vout (forwards to child vout)
 *****************************************************************************/
static int Control( vout_thread_t *p_vout, int i_query, va_list args )
{
    return vout_vaControl( p_vout->p_sys->p_vout, i_query, args );
}

/*****************************************************************************
 * SendEvents: forward mouse and keyboard events to the parent p_vout
 *****************************************************************************/
static int SendEvents( vlc_object_t *p_this, char const *psz_var,
                       vlc_value_t oldval, vlc_value_t newval, void *p_data )
{
    var_Set( (vlc_object_t *)p_data, psz_var, newval );

    return VLC_SUCCESS;
}
